#include <iostream>

#include "CppType.hpp"
#include "ParserWriter.hpp"
#include "XsdType.hpp"

using std::endl;
using std::string;
using boost::shared_ptr;

shared_ptr<CppType>
CppType::FromExternalType(const XsdType& xsdType)
{
    return shared_ptr<CppType>(new CppType(xsdType));
}

shared_ptr<CppType>
CppType::FromSimpleType(const string& typeName, const string& cppTypeName)
{
    return shared_ptr<CppType>(new CppType(typeName, cppTypeName));
}

CppType::CppType(const XsdType& xsdType):
    mpXsdType(&xsdType),
    mIsAnonymous(false),
    mIsSimple(false),
    mOutputName(mpXsdType->GetName())
{
    if (mOutputName.substr(0, 8) == "INTERNAL")
    {
        mIsAnonymous = true;
    }
    else
    {
        // User defined type.
        mOutputName[0] = toupper(mOutputName[0]);
        mOutputName = "Xml" + mOutputName;
    }
}

CppType::CppType(const string& typeName, const string& cppTypeName):
    mpXsdType(0),
    mIsAnonymous(false),
    mIsSimple(true),
    mOutputName(cppTypeName)
{
    // TODO: Do something with typeName?
}

string
CppType::GetHeaderFileName() const
{
    if (mIsAnonymous || mIsSimple)
    {
        return "";
    }
    else
    {
        return mOutputName + ".hpp";
    }
}

string
CppType::GetSourceFileName() const
{
    if (mIsAnonymous || mIsSimple)
    {
        return "";
    }
    else
    {
        return mOutputName + ".cpp";
    }
}

string
CppType::GetSimpleCastFromXmlChar(const string& argument) const
{
    assert(mIsSimple);

    if (mOutputName == "int")
        return "boost::lexical_cast<int>((char*)" + argument + ")";

    if (mOutputName == "double")
        return "boost::lexical_cast<double>((char*)" + argument + ")";

    if (mOutputName == "bool")
        return "(strcmp((char*)" + argument + ", \"true\") == 0) ? true : false";

    return "(char*)" + argument;
}

string
CppType::GetSimpleCastToXmlChar(const string& argument) const
{
    assert(mIsSimple);

    if (mOutputName == "int" || mOutputName == "double")
        return "BAD_CAST boost::lexical_cast<std::string>(" + argument + ").c_str()";

    if (mOutputName == "bool")
        return "BAD_CAST ((" + argument + " == true) ? \"true\" : \"false\")";

    if (mOutputName == "DATE"
           || mOutputName == "DATETIME"
           || mOutputName == "DURATION"
           || mOutputName == "TIME")
        return  "BAD_CAST " + argument + ".to_string().c_str()";

    return "BAD_CAST " + argument + ".c_str()";
}

void
CppType::GetIncludes(ParserWriter& context, std::set<string>& includes) const
{
    // TODO fix these includes to be more specific??
    includes.insert("\"XmlType.hpp\"");

    if (mIsSimple)
    {
        includes.insert("\"XmlSimpleTypeHandler.hpp\"");

        if (mOutputName == "std::string")
        {
            includes.insert("<string>");
        }
        else if (mOutputName == "DATE")
        {
            includes.insert("\"Date.hpp\"");
        }
        else if (mOutputName == "DATETIME")
        {
            includes.insert("\"DateTime.hpp\"");
        }
        else if (mOutputName == "DURATION")
        {
            includes.insert("\"Duration.hpp\"");
        }
        else if (mOutputName == "TIME")
        {
            includes.insert("\"Time.hpp\"");
        }
        else
        {
            includes.insert("<boost/lexical_cast.hpp>");
        }
    }
    else
    {
        mpXsdType->GetIncludes(context, includes);
    }
}

void
CppType::OutputAsClass(
        std::ostream& s, ParserWriter& context) const
{
    s << endl;
    s << "class " << mOutputName << ": public XmlType" << endl;
    s << "{" << endl;
    s << "public:" << endl;
    context.IncreaseIndent();
    s << context.GetIndent()
      << mOutputName << "(){};" << endl;
    s << context.GetIndent()
      << "~" << mOutputName << "(){};" << endl;

    if (mpXsdType->hasAttrFunction())
    {
        s << context.GetIndent()
          << "void DoAttribute(const xmlChar**);" << endl;
    }

    if (mpXsdType->hasDataFunction())
    {
        s << context.GetIndent()
          << "void DoData(const xmlChar*, int);" << endl;
    }

    if (mpXsdType->hasHandlerFunction())
    {
        s << context.GetIndent()
          << "boost::shared_ptr<XmlType> FindHandler(const xmlChar*);" << endl;
    }

#if GEN_DEBUG_CODE
    s << context.GetIndent()
      << "void Dump(xmlTextWriterPtr);" << endl;
#endif // GEN_DEBUG_CODE

    //TODO
    s << endl;
    mpXsdType->OutputGetters(s, context, 0);
    context.DecreaseIndent();
    s << endl;


    s << "private:" << endl;
    context.IncreaseIndent();
    mpXsdType->OutputClassMembers(s, context, 0);
    context.DecreaseIndent();
    s << endl;

    s << "};" << endl;
}

void
CppType::OutputAsCpp(
        std::ostream& s, ParserWriter& context) const
{
    s << endl;

    if (mpXsdType->hasAttrFunction())
    {
        // doAttribute
        s << context.GetIndent() << "void "
          << mOutputName << "::DoAttribute(const xmlChar** atts){" << endl;
        context.IncreaseIndent();
        // doAttribute contents
        s << context.GetIndent() << "if (!atts) return;" << endl;

        // while(*atts){
        s << context.GetIndent() << "while (*atts){" << endl;
        context.IncreaseIndent();
        // attribute functions
        s << context.GetIndent() << "if (false){}" << endl;
        mpXsdType->OutputAttrFunction(s, context);
        s << context.GetIndent() << "atts++;" << endl;
        // end while}
        context.DecreaseIndent();
        s << context.GetIndent() << "}" << endl;

        context.DecreaseIndent();
        s << context.GetIndent() << "}" << endl;
        s << endl;
        // end doAttribute
    }

    if (mpXsdType->hasDataFunction())
    {
        // doData (only valid for SimpleContent and extension elements)
        s << context.GetIndent() << "void "
          << mOutputName << "::DoData(const xmlChar* ch, int len){" << endl;
        context.IncreaseIndent();
        // doData contents
        mpXsdType->OutputDataFunction(s, context);
        context.DecreaseIndent();
        s << context.GetIndent() << "}" << endl;
        s << endl;
        // end doData
    }

    if(mpXsdType->hasHandlerFunction())
    {
        // findHandler
        s << context.GetIndent() << "boost::shared_ptr<XmlType> "
          << mOutputName << "::FindHandler(const xmlChar* name){" << endl;
        context.IncreaseIndent();
        // findHandler contents
        mpXsdType->OutputHandlerFunction(s, context, 0);
        s << context.GetIndent() << "return shared_from_this();" << endl;
        context.DecreaseIndent();
        s << context.GetIndent() << "}" << endl;
        s << endl;
        // end findHandler
    }

#if GEN_DEBUG_CODE
    s << context.GetIndent() << "void "
      << mOutputName << "::Dump(xmlTextWriterPtr w){" << endl;
    context.IncreaseIndent();
    mpXsdType->OutputDumpFunction(s, context, 0);
    context.DecreaseIndent();
    s << context.GetIndent() << "}" << endl;
    s << endl;
#endif // GEN_DEBUG_CODE
}